@@ -1,10 +1,10 @@
-__version__ = '1.0.2'
+__version__ = '1.2.0'
 
 from .config import *
 from .datasets import *
 from .dnnmodels import *
 from .evaluation import *
-from .plots import *
+from .run import *
 from .utils import *
 from .sensitivity import *
 from .surrogate import *
