@@ -7,6 +7,7 @@ import importlib
 # External
 import numpy
 import torch
+import matplotlib.pyplot as plt
 from torch.nn.parallel import DistributedDataParallel
     
 def train(itrial     = 0,
@@ -23,6 +24,7 @@ def train(itrial     = 0,
           log_dir    = 'logs',
           split      = 'data',
           rank       = 0,
+          transform  = None,
           **kwargs):
     """
     PyTorch training method. This will train built-in models available from HYPPO.
@@ -139,23 +141,24 @@ def train(itrial     = 0,
     logging.info('-'*40)
     logging.info('{} {:>3} / {:<3} {:>24}'.format('TRIAL',itrial+1,trial,'TESTING'))
     logging.info('-'*40)
-    test_loss = evaluate(device, data['test'], model, criterion, accuracy, **kwargs)
+    test_loss = evaluate(device, data['test'], model, criterion, accuracy, log_dir, output, **kwargs)
     # Save model if requested
+    checkpoint_dir = os.path.join(log_dir,'checkpoints',output[:-3])
+    state_path = os.path.join(checkpoint_dir,'%s.pth.tar' % os.path.basename(output))
     if split=='trial' or (split=='data' and rank==0):
-        checkpoint_dir = os.path.join(log_dir,'checkpoints')
-        path_to_model = os.path.join(checkpoint_dir,'%s.pth.tar' % os.path.basename(output))
         os.makedirs(checkpoint_dir, exist_ok=True)
-        torch.save(model.state_dict(),path_to_model)
-        model = str(path_to_model)
+        torch.save(model.state_dict(),state_path)
+    state_path = model
     return {
         'loss':test_loss.item(),
-        'models':model,
+        'loss_tf':test_loss.item() if transform==None else eval(transform).item(),
+        'models':state_path,
         'data':data,
         'criterion':criterion,
         'hyperprms':hyperprms,
     }
 
-def evaluate(device, data, model, criterion, accuracy=False, update=False, **kwargs):
+def evaluate(device, data, model, criterion, accuracy=False, log_dir=None, output=None, update=False, **kwargs):
     model.eval()
     start_time = time.time()
     logging.info('\t\tSize : {:>11}'.format(len(data.sampler)))
@@ -183,12 +186,16 @@ def evaluate(device, data, model, criterion, accuracy=False, update=False, **kwa
             if accuracy:
                 acc += get_accuracy(out,label)
         y_real = torch.tensor(y_real)
-#     import matplotlib.pyplot as plt
-#     plt.style.use('seaborn')
-#     plt.figure(figsize=(5,5))
-#     plt.plot(y_real.detach().numpy().squeeze(),color='blue')
-#     plt.plot(torch.stack(y_pred).detach().numpy().squeeze(),color='orange')
-#     plt.savefig('pred')
+    #if log_dir!=None and output!=None:
+    #    y_pred_to_save = torch.stack(y_pred).detach().numpy().squeeze()
+    #    path = os.path.join(log_dir,'output',output[:-3])
+    #    os.makedirs(path,exist_ok=True)
+    #    plt.style.use('seaborn')
+    #    plt.figure(figsize=(5,5))
+    #    plt.plot(y_real.detach().numpy().squeeze(),color='blue')
+    #    plt.plot(y_pred_to_save,color='orange')
+    #    plt.savefig(os.path.join(path,'y_pred_%s.pdf' % output[-2:]))
+    #    numpy.savetxt(os.path.join(path,'y_pred_%s.txt' % output[-2:]),y_pred_to_save,fmt='%f')
     loss = criterion(torch.stack(y_pred).to(device), y_real.to(device))
     logging.info('\t\tLoss : {:>11.5f}'.format(loss))
     if accuracy:
