@@ -6,6 +6,7 @@ import logging
 import importlib
 from collections import OrderedDict
 import numpy as np
+
 # External
 import torch
 import numpy
@@ -15,6 +16,22 @@ from scipy.integrate import quad
 import matplotlib.pyplot as plt
 
 def shape_and_size(data, n_classes=None, **kwargs):
+    """
+    Extract input/output shape and size. This function is used to determine
+    the right dimension of both input and output data.
+    
+    Parameters
+    ----------
+    data : :py:class:`dict`
+      Dictionary containing both input data (X_data key) and output labels (y_data key).
+    n_classes : :py:class:`int`
+      Number of output classes. Must be used when classification datasets are used.
+    
+    Returns
+    -------
+    prop : :py:class:`dict`
+      Input/output shape and size dimensions.
+    """
     # Get input and output datasets
     X_data, y_data = next(iter(data))
     # Extract input/output size and shapes
@@ -26,6 +43,16 @@ def shape_and_size(data, n_classes=None, **kwargs):
     return prop
 
 def get_fct(category,fct=None):
+    """
+    Load PyTorch function of target element.
+    
+    Parameters
+    ----------
+    category : :py:class:`str`
+      Target type of function to lookup in PyTorch (i.e. activation, loss, or optimizer function).
+    fct : :py:class:`str`
+      Target function to lookup with the above category.
+    """
     pytorch_dict = {
         # Activation functions,
         'activation':{
@@ -70,6 +97,160 @@ def get_fct(category,fct=None):
     }
     return pytorch_dict[category] if fct==None else pytorch_dict[category][fct]
 
+def run_uq(output, log_dir, data, models, hyperprms, loss, dl_type, trial, **kwargs):
+    """
+    Execute Monte-Carlo Dropout Algorithm. This function loads the trained model
+    architecture and does uncertainty quantification using MC dropout technique.
+    
+    Parameters
+    ----------
+    output : :py:class:`str`
+      Path to model output directory where trained models are saved. This has a unique
+      name that contains the value of each hyperparameter.
+    log_dir : :py:class:`str`
+      Path to parent repository where all output directories are stored.
+    data : :py:class:`dict`
+      Loaded training, validation and test datasets.
+    uq : :py:class:`dict`
+      Uncertainty quantification parameter settings. This corresponds to the
+      UQ information provided by the user in the configuration file (
+      :ref:`see here <uq>` for more details).
+    """
+    device = torch.device('cpu')
+    # Architecture initialization
+    module = importlib.import_module('.' + dl_type.lower(), 'hyppo.dnnmodels.pytorch')
+    net = module.get_model(data=data['train'], prms=hyperprms, **kwargs).to(device)
+    trained_models = []
+    for i,model in enumerate(models):
+        if type(model)==numpy.str_:
+            trained_models.append(load_model(model,net,device))
+        else:
+            trained_models.append(model)
+    dropout_rate = hyperprms.get('dropout',0.05)[0]
+    if dropout_rate == 0:
+        logging.warning('Running uncertainty quantification with dropout rate 0 is not optimal.')
+        logging.warning('Dropout rate set to 5% for Monte Carlo dropout UQ analysis as a default.')
+    pred_mean, total_var, confidence_interval = uq_quantifier(output, log_dir, data, trained_models, **kwargs)
+    logging.info('-'*40)
+    logging.info('OUTER OBJECTIVE FUNCTION {:>8} TRIALS'.format(len(loss)))
+    logging.info('-'*40)
+    logging.info('\tOuter Loss           : {:>9.5f}'.format(confidence_interval[0]))
+    logging.info('\tLower Bound of CI    : {:>9.5f}'.format(confidence_interval[1]))
+    logging.info('\tUpper Bound of CI    : {:>9.5f}'.format(confidence_interval[2]))
+    return confidence_interval[0] #{'pred_mean':pred_mean, 'pred_var': total_var,'confidence_int':confidence_interval}
+  
+def load_model(fname,model,device='cpu'):
+    """
+    Load saved model's parameter dictionary to initialized model.
+    The function will remove any ``.module`` string from parameter's name.
+
+    Parameters
+    ----------
+    fname : :py:class:`str`
+      Path to saved model
+    model : :py:class:`torch.nn.Module`
+      Initialized network network architecture
+    device : :py:class:`str`
+      Type of processor used (CPU or GPU'
+
+    Returns
+    -------
+    model : :py:class:`torch.nn.Module`
+      Up-to-date neural network model
+    """
+    state_dict = torch.load(fname, map_location=device)
+    new_state_dict = OrderedDict()
+    for key, value in state_dict.items():
+        if key.startswith('module.'):
+            key = key[7:]
+        new_state_dict[key] = value
+    model.load_state_dict(new_state_dict)
+    return model
+
+def uq_quantifier(output, log_dir, data, trained_models, criterion, uq_weights=[0.5,0.5],
+                  dropout_masks=30, data_noise=0, data_type=None, update=False, device=None, **kwargs):
+    data_variance = data_noise
+    weights = uq_weights
+    T = len(trained_models)
+    alt_models=[]
+    if data_variance == 0.0:
+        for i in range(T):
+            alt_models.append(Dropout_Model_Builder(trained_models[i]))
+            # fake_model = trained_models[i]
+            # alt_models.append(fake_model)
+    else:
+        for i in range(T):
+            alt_models.append(ADF_Dropout_Model_Builder(trained_models[i],data_variance))
+    uq_means = {'trained':{}, 'Unc':{}}
+    uq_vars  = {'trained':{}, 'Unc':{}}
+    uq_obj_vals = []
+    N = dropout_masks
+    logging.info('-'*40)
+    logging.info('UNCERTAINTY QUANTIFICATION {:>6} MODELS'.format(T))
+    logging.info('-'*40)
+    logging.info('Model | Obj. Value | Dropout Mask Time')
+    for j in range(T):
+        if data_variance > 0.0:
+            model = ADF_Dropout_Model_Builder(trained_models[j],data_variance,True)
+        else:
+            model = trained_models[j]
+            model.eval()
+        trained_val, trained_pred, trained_var, y_real = pred_evaluate(data['test'], model, criterion,update,device,**kwargs)
+        uq_obj_vals.append(trained_val)
+        uq_means['trained'][j] = numpy.asarray(trained_pred)
+        uq_vars['trained'][j]  = numpy.asarray(trained_var)
+        start_time = time.time()
+        model = alt_models[j]
+        model.train()
+        for t in range(N): 
+            val, pred, pred_var, y_real= pred_evaluate(data['test'], model, criterion, update, device, **kwargs)
+            uq_obj_vals.append(val)
+            uq_means['Unc'][j*N+t] = numpy.asarray(pred)
+            uq_vars['Unc'][j*N+t] = numpy.asarray(pred_var)
+        logging.info('{:>5} | {:>10.5f} | {:>17.3f} s'.format(j+1,trained_val.detach(),time.time()-start_time))
+    pred_mean = (weights[0]/T)*sum(uq_means['trained'].values())+weights[1]*sum(uq_means['Unc'].values())/(T*N)
+    data_var  = (weights[0]/T)*sum(uq_vars['trained'].values())+weights[1]*sum(uq_vars['Unc'].values())/(T*N)
+    model_trained_var = torch.square(torch.tensor(uq_means['trained'][0]) - torch.tensor(pred_mean)) 
+    model_Unc_var     = torch.square(torch.tensor(uq_means['Unc'][0]) - torch.tensor(pred_mean)) 
+    for i  in range(1,T):
+        model_trained_var = model_trained_var + torch.square(torch.tensor(uq_means['trained'][i]) - torch.tensor(pred_mean))
+    for j in range(1,T*N):
+        model_Unc_var = model_Unc_var + torch.square(torch.tensor(uq_means['Unc'][j]) - torch.tensor(pred_mean)) 
+    model_var = (weights[0]/T)*numpy.asarray(model_trained_var)+ (weights[1]/(N*T))*numpy.asarray(model_Unc_var)
+    total_var = data_var + model_var
+    pred_mean_obj_val = criterion(torch.tensor(pred_mean),y_real)
+    pred_std = numpy.std(torch.stack(uq_obj_vals).cpu().detach().numpy())
+    if data_type=='im':
+        path = os.path.join(log_dir,'output',output)
+        os.makedirs(path,exist_ok=True)
+        numpy.savetxt(os.path.join(path,'mean.txt'),pred_mean,fmt='%f')
+        numpy.savetxt(os.path.join(path,'var.txt'),model_var,fmt='%f')
+        numpy.savetxt(os.path.join(path,'y_real.txt'),y_real,fmt='%i')
+    elif data_type=='ts':
+        path = os.path.join(log_dir,'output',output)
+        os.makedirs(path,exist_ok=True)
+        trained_means = numpy.array([uq_means['trained'][key].squeeze() for key in uq_means['trained'].keys()]).T
+        uq_means = numpy.array([uq_means['Unc'][key].squeeze() for key in uq_means['Unc'].keys()]).T
+        numpy.savetxt(os.path.join(path,'trained_means.txt'),trained_means,fmt='%f')
+        numpy.savetxt(os.path.join(path,'uq_means.txt'),uq_means,fmt='%f')
+        numpy.savetxt(os.path.join(path,'pred_mean.txt'),pred_mean.squeeze(),fmt='%f')
+        numpy.savetxt(os.path.join(path,'pred_var.txt'),total_var.squeeze(),fmt='%f')
+
+    #print('uq_obj_val={}'.format(numpy.asarray(uq_obj_vals)))
+    #print('pred_mean_obj_val  ={}'.format(pred_mean_obj_val))
+    #print('confidence intevals = [{},{},{}]'.format(pred_mean_obj_val-2*pred_std, pred_mean_obj_val, pred_mean_obj_val+2*pred_std))
+    #print('pred_mean = {}'.format(pred_mean.squeeze()))
+    #print('total var = {}'.format(total_var.squeeze()))
+    #print('number correct:{}'.format(numpy.count_nonzero((numpy.asarray(y_real)-numpy.argmax(numpy.asarray(uq_means['trained'][0]),axis=1))==0)))
+    #print('trained model pred={}'.format(numpy.argmax(numpy.asarray(uq_means['trained'][0]),axis=1)))
+
+    # Return statistics
+    lb = pred_mean_obj_val-2*pred_std
+    ub = pred_mean_obj_val+2*pred_std
+    confidence_interval = [pred_mean_obj_val,lb,ub]
+    
+    return pred_mean, total_var, confidence_interval
+
 ############################################################################################
 # These classes construct the layers which propogate data uncertianty using ADF
 #-------------------------------------------------------------------------------------------
@@ -150,15 +331,14 @@ def adf_layer_selector(model, lay_num, lay_name, dropout_rate):
     else:
         layer = ADF_Other_Act_Layer(lay_name)
     return layer
-        
 
 ############################################################################################
 class ADF_Dropout_Model_Builder(torch.nn.Module):
     
-    def __init__(self, in_shape, out_shape, model, data_noise, exact = False):
+    def __init__(self, model, data_noise, exact = False):
         super(ADF_Dropout_Model_Builder, self).__init__()
-        self.in_size  = numpy.prod(in_shape)
-        self.out_shape = out_shape
+        self.in_size = model.prop['input_size']
+        self.out_shape = model.prop['output_shape']
         self.noise = data_noise
         lay_names = []
         for i in range(len(model.layers)):
@@ -202,8 +382,10 @@ class ADF_Dropout_Model_Builder(torch.nn.Module):
 ############################################################################################
 class Dropout_Model_Builder(torch.nn.Module):
 
-    def __init__(self, model):
+    def __init__(self, model, **kwargs):
         super(Dropout_Model_Builder, self).__init__()
+        self.in_size = model.prop['input_size']
+        self.out_shape = model.prop['output_shape']
         lay_names = []
         for i in range(len(model.layers)):
             lay_names.append(repr(type(model.layers[i])).split(".")[-1].split("'")[0])
@@ -230,19 +412,19 @@ class Dropout_Model_Builder(torch.nn.Module):
                 else:
                     layers.append(model.layers[i])
         self.layers = torch.nn.Sequential(*layers)
-        
+
     def forward(self,data):
         if self.lay_names[0] == 'Linear':
-            data = data.reshape(data.shape[0],self.in_model.prop['input_size'])
+            data = data.reshape(data.shape[0],self.in_size) #This will need to be update to use size and shape function
             out = self.layers(data)
-            out = out.reshape(data.shape[0],*self.in_model.prop['output_shape'])
-            return out
+            out = out.reshape(data.shape[0],*self.out_shape)
         elif self.lay_names[0] == 'Conv2d':
-            return 0 
+            out = self.layers(data)
+            out = out.reshape(data.shape[0],*self.out_shape)
         elif self.lay_names[0] == 'LSTM':
-            return self.layers(data)
+            out =  self.layers(data)
+        return out
 
-############################################################################################
 def pred_evaluate(data, model, loss_function, update=False, device=None, **kwargs):
     device = torch.device('cpu')
     model = model.to(device)
@@ -281,10 +463,9 @@ def pred_evaluate(data, model, loss_function, update=False, device=None, **kwarg
         y_real = torch.stack(y_real)
     y_pred_mean = torch.stack(y_pred_mean)
     loss = loss_function(y_pred_mean.to(device), y_real.to(device))
-    return loss, y_pred_mean.cpu().detach().numpy(), torch.tensor(y_pred_var), y_real
+    return loss, y_pred_mean.cpu().detach().numpy(), torch.tensor(numpy.asarray(y_pred_var)), y_real
 
-############################################################################################
-def time_series_uq_plotter(data, trained_means, pred_mean, pred_var, uq_means, plot_title):
+def time_series_uq_plotter(output, data, trained_means, pred_mean, pred_var, uq_means, plot_title):
     train_data   = data['train']
     valid_data   = data['valid']
     test_data    = data['test']
@@ -319,239 +500,5 @@ def time_series_uq_plotter(data, trained_means, pred_mean, pred_var, uq_means, p
     plt.ylabel('Normalized Temperature')
     plt.tight_layout()
     plt.legend(loc='best')
-    plt.savefig(os.path.join('plot_data/time series',plot_title+'.pdf'))
+    plt.savefig(os.path.join('logs/output',output,'time_series',plot_title+'.pdf'))
     plt.close()
-# ############################################################################################
-def image_classification_plot(mean, var, correct_class, plt_name):
-    max_val = numpy.max(mean + numpy.sqrt(var))
-    min_val = numpy.min(mean - numpy.sqrt(var))
-    diff    = max_val - min_val
-    scaled_mean_std = (mean+numpy.sqrt(var)-min_val)/diff
-    scaled_mean = (mean - min_val)/diff
-    scaled_std  = scaled_mean_std - scaled_mean
-    os.makedirs('plot_data/image',exist_ok=True)
-    plt.figure(figsize=(12,7),dpi=250)
-    indx = numpy.linspace(1,len(mean),len(mean))
-    plt.style.use('seaborn-whitegrid')
-    plt.errorbar(indx, scaled_mean, scaled_std, fmt='.r', ms=10, mfc = 'k', capsize=5.0)
-    plt.rc('font',size=25)
-    plt.xlabel('Class',fontweight='bold',fontsize=24)
-    plt.ylabel('Prediction Score',fontweight='bold',fontsize=24)
-    plt.ylim([-0.1,1.1])
-    plt_title = 'Image Classification - Correct Class '+str(correct_class)+', Predicted Class '+str(numpy.argmax(mean)+1)
-    #plt.legend(['Prediction Score','Scaled Prediction Score'])
-    #plt.errorbar(indx, mean, numpy.sqrt(var),fmt = '.b', ms=3, mfc = 'k',capsize = 5.0)
-    plt.title(plt_title,fontweight = 'bold',fontsize=24)
-    plt.savefig(os.path.join('plot_data/image',plt_name))
-    plt.close()
-
-############################################################################################
-def uq_quantifer(data, trained_models, model_prms, loss_function, data_variance, weights, update=False, plot_type=None, device=None, **kwargs):
-    T = len(trained_models)
-    alt_models=[]
-    if data_variance == 0.0:
-        for i in range(T):
-#             alt_models.append(Dropout_Model_Builder(trained_models[i],**model_prms))
-            fake_model = trained_models[i]
-            alt_models.append(fake_model)
-    else:
-        for i in range(T):
-            alt_models.append(ADF_Dropout_Model_Builder(model_prms['input_shape'],model_prms['output_shape'],trained_models[i],data_variance))
-    uq_means = {'trained':{}, 'Unc':{}}
-    uq_vars  = {'trained':{}, 'Unc':{}}
-    uq_obj_vals = []
-    N = 1
-    logging.info('-'*40)
-    logging.info('UNCERTAINTY QUANTIFICATION {:>6} MODELS'.format(T))
-    logging.info('-'*40)
-    logging.info('Model | Obj. Value | Dropout Mask Time')
-    for j in range(T):
-        if data_variance > 0.0:
-            model = ADF_Dropout_Model_Builder(model_prms['input_shape'],model_prms['output_shape'],trained_models[j],data_variance,True)
-        else:
-            model = trained_models[j]
-            model.train()
-        trained_val, trained_pred, trained_var, y_real = pred_evaluate(data['test'], model, loss_function,update,device,**kwargs)
-        uq_obj_vals.append(trained_val)
-        uq_means['trained'][j] = numpy.asarray(trained_pred)
-        uq_vars['trained'][j]  = numpy.asarray(trained_var)
-        start_time = time.time()
-        model = alt_models[j]
-        model.train()
-        for t in range(N): 
-            val, pred, pred_var, y_real= pred_evaluate(data['test'], model, loss_function, update, device, **kwargs)
-            uq_obj_vals.append(val)
-            uq_means['Unc'][j*N+t] = numpy.asarray(pred)
-            uq_vars['Unc'][j*N+t] = numpy.asarray(pred_var)
-        logging.info('{:>5} | {:>10.5f} | {:>17.3f} s'.format(j+1,trained_val.detach(),time.time()-start_time))
-    pred_mean = (weights[0]/T)*sum(uq_means['trained'].values())+weights[1]*sum(uq_means['Unc'].values())/(T*N)
-    data_var  = (weights[0]/T)*sum(uq_vars['trained'].values())+weights[1]*sum(uq_vars['Unc'].values())/(T*N)
-    model_trained_var = torch.square(torch.tensor(uq_means['trained'][0]) - torch.tensor(pred_mean)) 
-    model_Unc_var     = torch.square(torch.tensor(uq_means['Unc'][0]) - torch.tensor(pred_mean)) 
-    for i  in range(1,T):
-        model_trained_var = model_trained_var + torch.square(torch.tensor(uq_means['trained'][i]) - torch.tensor(pred_mean))
-    for j in range(1,T*N):
-        model_Unc_var = model_Unc_var + torch.square(torch.tensor(uq_means['Unc'][j]) - torch.tensor(pred_mean)) 
-    model_var = (weights[0]/T)*numpy.asarray(model_trained_var)+ (weights[1]/(N*T))*numpy.asarray(model_Unc_var)
-    total_var = data_var + model_var
-    pred_mean_obj_val = loss_function(torch.tensor(pred_mean),y_real)
-    pred_std = numpy.std(torch.stack(uq_obj_vals).cpu().detach().numpy())
-    print(plot_type)
-    if plot_type=='image':
-        for i in range(2):
-            image_classification_plot(pred_mean[i,:], model_var[i,:], numpy.asarray(y_real)[i]+1,'Test '+str(i))
-    elif plot_type=='time_series':
-        time_series_uq_plotter(data,uq_means['trained'],pred_mean,total_var,uq_means['Unc'],'Test')
-
-    #print('uq_obj_val={}'.format(numpy.asarray(uq_obj_vals)))
-    #print('pred_mean_obj_val  ={}'.format(pred_mean_obj_val))
-    #print('confidence intevals = [{},{},{}]'.format(pred_mean_obj_val-2*pred_std, pred_mean_obj_val, pred_mean_obj_val+2*pred_std))
-    #print('pred_mean = {}'.format(pred_mean.squeeze()))
-    #print('total var = {}'.format(total_var.squeeze()))
-    #print('number correct:{}'.format(numpy.count_nonzero((numpy.asarray(y_real)-numpy.argmax(numpy.asarray(uq_means['trained'][0]),axis=1))==0)))
-    #print('trained model pred={}'.format(numpy.argmax(numpy.asarray(uq_means['trained'][0]),axis=1)))
-
-    # Return statistics
-    lb = pred_mean_obj_val-2*pred_std
-    ub = pred_mean_obj_val+2*pred_std
-    confidence_interval = [pred_mean_obj_val,lb,ub]
-    
-    return pred_mean, total_var, confidence_interval
-
-############################################################################################
-def run_uq(data, criterion, models, hyperprms, loss, dl_type, trial, uq={}, **kwargs):
-    device = torch.device('cpu')
-    # Architecture initialization
-    module = importlib.import_module('.' + dl_type.lower(), 'hyppo.dnnmodels.pytorch')
-    net = module.get_model(data=data['train'], prms=hyperprms, **kwargs).to(device)
-    classes = []
-    for i,model in enumerate(models):
-        if type(model)==numpy.str_:
-            classes.append(load_model(model,net,device))
-    if len(classes)>0:
-        models = classes
-    data_shape = shape_and_size(data['test'], **kwargs)
-    data_variance = uq.get('data_noise',0.0)
-    weights = uq.get('uq_weights', [0.5,0.5])
-    dropout_rate = hyperprms.get('dropout',0.05)[0]
-    if dropout_rate == 0:
-        logging.warning('Running uncertainty quantification with dropout rate 0 is not optimal.')
-        logging.warning('Dropout rate set to 5% for Monte Carlo dropout UQ analysis as a default.')
-    pred_mean, total_var, confidence_interval = uq_quantifer(data, models, data_shape, criterion, data_variance, weights, **kwargs)
-    logging.info('-'*40)
-    logging.info('OUTER OBJECTIVE FUNCTION {:>10} TRIALS'.format(len(loss)))
-    logging.info('-'*40)
-    logging.info('\tMedian Loss (pre-UQ) : {:>9.5f}'.format(numpy.nanmedian(loss)))
-    logging.info('\tMedian Devation      : {:>9.5f}'.format(scipy.stats.median_absolute_deviation(loss,scale='normal')))
-    logging.info('\tOuter Loss           : {:>9.5f}'.format(confidence_interval[0]))
-    logging.info('\tLower Bound of CI    : {:>9.5f}'.format(confidence_interval[1]))
-    logging.info('\tUpper Bound of CI    : {:>9.5f}'.format(confidence_interval[2]))
-    return confidence_interval[0] #{'pred_mean':pred_mean, 'pred_var': total_var,'confidence_int':confidence_interval}
-    
-def load_model(fname,model,device='cpu'):
-    """
-    Load saved model's parameter dictionary to initialized model.
-    The function will remove any ``.module`` string from parameter's name.
-
-    Parameters
-    ----------
-    fname : :py:class:`str`
-      Path to saved model
-    model : :py:class:`torch.nn.Module`
-      Initialized network network architecture
-
-    Returns
-    -------
-    model : :py:class:`torch.nn.Module`
-      Up-to-date neural network model
-    """
-    state_dict = torch.load(fname, map_location=device)
-    new_state_dict = OrderedDict()
-    for key, value in state_dict.items():
-        if key.startswith('module.'):
-            key = key[7:]
-        new_state_dict[key] = value
-    model.load_state_dict(new_state_dict)
-    return model
-
-
-def uq_quantifer(data, trained_models, model_prms, loss_function, data_variance, weights, update=True, plot_type=None):
-    T = len(trained_models)
-    alt_models=[]
-    if data_variance == 0.0:
-        for i in range(T):
-            alt_models.append(Dropout_Model_Builder(trained_models[i]))
-    else:
-        for i in range(T):
-            alt_models.append(ADF_Dropout_Model_Builder(model_prms['in_shape'],model_prms['out_shape'],trained_models[i],data_variance))
-    uq_means = {'trained':{}, 'Unc':{}}
-    uq_vars  = {'trained':{}, 'Unc':{}}
-    uq_obj_vals = []
-    N = 15
-    for j in range(T):
-        if data_variance > 0.0:
-            model = ADF_Dropout_Model_Builder(model_prms['in_shape'],model_prms['out_shape'],trained_models[j],data_variance,True)
-        else:
-            model = trained_models[j]
-            model.eval()
-        trained_val, trained_pred, trained_var, y_real = pred_evaluate(data['test'], model, loss_function,update)
-        uq_obj_vals.append(trained_val)
-        uq_means['trained'][j] = np.asarray(trained_pred)
-        uq_vars['trained'][j]  = np.asarray(trained_var)
-        print('Trained Model {} Obj. Val:{}'.format(j+1,trained_val.detach()))
-        print('Dropout Masks for Trained Model {}'.format(j+1))
-        for t in range(N): 
-            val, pred, pred_var, y_real= pred_evaluate(data['test'], alt_models[j], loss_function, update)
-            uq_obj_vals.append(val)
-            uq_means['Unc'][j*N+t] = np.asarray(pred)
-            uq_vars['Unc'][j*N+t] = np.asarray(pred_var)
-    pred_mean = (weights[0]/T)*sum(uq_means['trained'].values())+weights[1]*sum(uq_means['Unc'].values())/(T*N)
-    data_var  = (weights[0]/T)*sum(uq_vars['trained'].values())+weights[1]*sum(uq_vars['Unc'].values())/(T*N)
-    model_trained_var = torch.square(torch.tensor(uq_means['trained'][0]) - torch.tensor(pred_mean)) 
-    model_Unc_var     = torch.square(torch.tensor(uq_means['Unc'][0]) - torch.tensor(pred_mean)) 
-    for i  in range(1,T):
-        model_trained_var = model_trained_var + torch.square(torch.tensor(uq_means['trained'][i]) - torch.tensor(pred_mean)) 
-    for j in range(1,T*N):
-        model_Unc_var = model_Unc_var + torch.square(torch.tensor(uq_means['Unc'][j]) - torch.tensor(pred_mean)) 
-    model_var = (weights[0]/T)*np.asarray(model_trained_var)+ (weights[1]/(N*T))*np.asarray(model_Unc_var)
-    total_var = data_var + model_var
-    pred_mean_obj_val = loss_function(torch.tensor(pred_mean),y_real) 
-    pred_std = np.std(torch.stack(uq_obj_vals).cpu().detach().numpy())
-
-    if plot_type=='image':
-        for i in range(2):
-            image_classification_plot(pred_mean[i,:], model_var[i,:], np.asarray(y_real)[i]+1,'Test '+str(i))
-    elif plot_type=='time_series':
-            time_series_uq_plotter(data,uq_means['trained'],pred_mean,total_var,uq_means['Unc'],'Test')
-
-    #print('uq_obj_val={}'.format(np.asarray(uq_obj_vals)))
-    #print('pred_mean_obj_val  ={}'.format(pred_mean_obj_val))
-    #print('confidence intevals = [{},{},{}]'.format(pred_mean_obj_val-2*pred_std, pred_mean_obj_val, pred_mean_obj_val+2*pred_std))
-    #print('pred_mean = {}'.format(pred_mean.squeeze()))
-    #print('total var = {}'.format(total_var.squeeze()))
-    #print('number correct:{}'.format(np.count_nonzero((np.asarray(y_real)-np.argmax(np.asarray(uq_means['trained'][0]),axis=1))==0)))
-    #print('trained model pred={}'.format(np.argmax(np.asarray(uq_means['trained'][0]),axis=1)))
-
-    # Return statistics
-    lb = pred_mean_obj_val-2*pred_std
-    ub = pred_mean_obj_val+2*pred_std
-    confidence_interval = [pred_mean_obj_val,lb,ub]
-  
-    return pred_mean, total_var, confidence_interval
-
-############################################################################################
-def run_uq(data, criterion, models, hyperprms, loss, update, uq={}, **kwargs):
-    trained_models = []
-    #for model in 
-    #print(models[0],type(models[0]))
-    data_shape = shape_and_size(data['test'], **kwargs)
-    data_variance = uq.get('data_noise',0.0)
-    weights = uq.get('uq_weights', [0.5,0.5])
-#    update  = uq['update']
-    dropout_rate = hyperprms.get('dropout',0.05)[0]
-    if dropout_rate == 0:
-        logging.warning('Running uncertainty quantification with dropout rate 0 is not optimal.')
-        logging.warning('Dropout rate set to 5% for Monte Carlo dropout UQ analysis as a default.')
-    pred_mean, total_var, confidence_interval = uq_quantifer(data, models, data_shape, criterion, data_variance, weights,update)
-    return {'pred_mean':pred_mean, 'pred_var': total_var,'confidence_int':confidence_interval}
-
