@@ -1,15 +1,15 @@
 # System
 import os
 import sys
-import pickle
 import logging
 import importlib
 import time
+import types
 
 # External
 import yaml
 import numpy
-from scipy.stats import median_abs_deviation
+import pickle5 as pickle
 
 # Local
 from .datasets import get_loader
@@ -18,7 +18,6 @@ from .hyperparams import get_hyperprms
 from .utils.extract import gather_trials, check_evaluation_outer_loss
 from .obj_fct import *
 
-
 def train_evaluation(x_sc, samples, trainer, model, dist, output='out', data={}, uq={}, **kwargs):
     """
     Execute training according to requested trainer mode. This function looks at the
@@ -30,15 +29,20 @@ def train_evaluation(x_sc, samples, trainer, model, dist, output='out', data={},
     x_sc : :class:`dict`
       Random hyperparameter set. Each key corresponds to the hyperparameter names
       and the corresponding value to the randomly selected sample value.
+    samples : :class:`numpy.ndarray`
+      Random hyperparameter integer values for single target evaluation.
     trainer : :class:`str`
       Trainer mode to be used.
-    data : :class:`dict`
-      Data setting extracted from the YAML configuration file.
     model : :class:`dict`
       Model setting extracted from the YAML configuration file.
     output : :class:`str`
       Directory name to save output training results.
-      
+    data : :class:`dict`
+      Data setting extracted from the YAML configuration file (empty by default).
+    uq : :class:`dict`
+      Uncertainty quantification parameter setting extracted from the YAML configuration
+      file (empty by default).
+
     Returns
     -------
     res : :class:`float`
@@ -47,7 +51,7 @@ def train_evaluation(x_sc, samples, trainer, model, dist, output='out', data={},
     # Start timer for recording function evaluation time
     start_time = time.time()
     # Display hyperparameter set only once, before the first trial
-    hyperprms = get_hyperprms(x_sc, **model)
+    hyperprms = get_hyperprms(trainer, x_sc, **model)
     # Prepare list of trial indexes to be evaluated in rank
     trials = numpy.arange(model['trial'])
     if dist['split']=='trial':
@@ -55,17 +59,18 @@ def train_evaluation(x_sc, samples, trainer, model, dist, output='out', data={},
     # Loop through all selected trials
     all_results = []
     for i in trials:
-        out_path = '%s-%02i' % (output,i+1)
+        out_path = '%s/%02i' % (output,i+1)
         if trainer=='internal':
             loaders = get_loader(data, **hyperprms, **model, **dist)
             results = training(itrial=i,data=loaders,hyperprms=hyperprms,output=out_path,**model,**dist)
         else:
-            results = external_training(i, x_sc, trainer, out_path, model)
+            results = external_training(i, x_sc, trainer, out_path, model, **dist)
         all_results.append(results)
     # Prepare single results dictionary
     results = all_results[0]
-    results['loss'] = [all_results[n]['loss'] for n in range(len(all_results))]
-    results['models'] = [all_results[n]['models'] for n in range(len(all_results))]
+    results['output'] = output
+    for key in ['loss','loss_tf','models']:
+        results[key] = [model_results[key] for model_results in all_results]
     # Check if all losses over all trials are found
     res = None
     if dist['rank']==0:
@@ -75,47 +80,47 @@ def train_evaluation(x_sc, samples, trainer, model, dist, output='out', data={},
             results = gather_trials(results,samples,out_path,**model,**dist)
         # Remove trials that gave nan values
         idxs = numpy.argwhere(numpy.isnan(results['loss']))
-        results['loss'] = numpy.delete(results['loss'],idxs)
-        results['models'] = numpy.delete(results['models'],idxs)
+        for key in ['loss','loss_tf','models']:
+            results[key] = numpy.delete(results[key],idxs)
         # Execute UQ or not
         if 'uq_on' in uq.keys() and uq['uq_on']==True:
-            res = run_uq(**results,**model,**dist)
+            res = run_uq(**results,**model,**dist,**uq)
         else:
             logging.info('-'*40)
             logging.info('OUTER OBJECTIVE FUNCTION {:>8} TRIALS'.format(len(results['loss'])))
             logging.info('-'*40)
-            logging.info('\tOuter Loss : {:>11.5f}'.format(numpy.median(results['loss'])))
-            logging.info('\tMAD        : {:>11.5f}'.format(median_abs_deviation(results['loss'],scale='normal')))
-            logging.info('\tMean Loss  : {:>11.5f}'.format(numpy.mean(results['loss'])))
+            logging.info('\tOuter Loss : {:>11.5f}'.format(numpy.mean(results['loss_tf'])))
+            logging.info('\tInner Loss : {:>11.5f}'.format(numpy.mean(results['loss'])))
             logging.info('\tSTD        : {:>11.5f}'.format(numpy.std(results['loss'])))
-            res = numpy.nanmedian(results['loss'])
+            res = numpy.mean(results['loss_tf'])
     else:
         logging.info('-'*40)
         logging.info('Waiting for other processes to finish...')
         while check_evaluation_outer_loss(samples,out_path,**dist)==False:
             continue
     logging.info('-'*40)
-    logging.info("Execution Time : %.3f s" % (time.time() - start_time))
+    logging.info("Execution Time : %.3f s" % (time.time() - start_time)) 
     return res
 
-def external_training(itrial, x_sc, trainer, output_dir, model):
+def external_training(itrial, x_sc, trainer, output_dir, model, log_dir, **kwargs):
     start_time = time.time()
     logging.info('-'*40)
     logging.info('{} {:>3} / {:<3} {:>24}'.format('TRIAL',itrial+1,model['trial'],'TESTING'))
     logging.info('-'*40)
-    if 'output_dir' in model.keys():
-        model['output_dir'] = output_dir
+    model['output_dir'] = os.path.join(log_dir,'output',output_dir)
     if 'yaml_input' in model.keys():
-        params = yaml_creation(x_sc,output_dir,**model['yaml_input'])
+        params = yaml_creation(x_sc,model['output_dir'],**model['yaml_input'])
     else:
         params = {**model,**x_sc}
-    modlist = trainer.split('.')
-    module = importlib.import_module('.' + modlist[-2], '.'.join(modlist[:-2]))
-    loss = eval('module.'+modlist[-1])(**params)
-    print(loss)
+    if type(trainer)==types.FunctionType:
+        loss = trainer(**params)
+    else:
+        modlist = trainer.split('.')
+        module = importlib.import_module('.' + modlist[-2], '.'.join(modlist[:-2]))
+        loss = eval('module.'+modlist[-1])(**params)
     logging.info('\t\tLoss : {:>11.5f}'.format(loss))
     logging.info('\t\tTime : {:>11.5f} s'.format(time.time()-start_time))
-    return {'loss':loss,'models':None}
+    return {'loss':loss,'loss_tf':loss,'models':None}
 
 def yaml_creation(x_sc,output_dir,src,config=None,**kwargs):
     # Extract input configuration file
@@ -132,4 +137,4 @@ def yaml_creation(x_sc,output_dir,src,config=None,**kwargs):
     with open(yaml_config, 'w') as f:
         yaml.dump(params, f, default_flow_style=False)
     params = {'yaml_config':yaml_config,'config':config,**kwargs}
-    return params
\ No newline at end of file
+    return params
